/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Shell Service.
 *
 * The Initial Developer of the Original Code is mozilla.org.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsKDEShellService.h"
#include "nsShellService.h"
#include "nsKDEUtils.h"
#include "nsCOMPtr.h"
#include "nsIPrefService.h"
#include "nsIProcess.h"
#include "nsILocalFile.h"
#include "nsServiceManagerUtils.h"
#include "nsComponentManagerUtils.h"

nsresult
nsKDEShellService::Init()
    {
    if( !nsKDEUtils::kdeSupport())
        return NS_ERROR_NOT_AVAILABLE;
    return NS_OK;
    }

NS_IMPL_ISUPPORTS1(nsKDEShellService, nsIShellService)

NS_IMETHODIMP
nsKDEShellService::IsDefaultBrowser(PRBool aStartupCheck,
                                      PRBool* aIsDefaultBrowser)
    {
    *aIsDefaultBrowser = PR_FALSE;
    if (aStartupCheck)
        mCheckedThisSession = PR_TRUE;
    nsCStringArray command;
    command.AppendCString( NS_LITERAL_CSTRING( "ISDEFAULTBROWSER" ));
    if( nsKDEUtils::command( command ))
        *aIsDefaultBrowser = PR_TRUE;
    return NS_OK;
    }

NS_IMETHODIMP
nsKDEShellService::SetDefaultBrowser(PRBool aClaimAllTypes,
                                       PRBool aForAllUsers)
    {
    nsCStringArray command;
    command.AppendCString( NS_LITERAL_CSTRING( "SETDEFAULTBROWSER" ));
    command.AppendCString( aClaimAllTypes ? NS_LITERAL_CSTRING( "ALLTYPES" ) : NS_LITERAL_CSTRING( "NORMAL" ));
    return nsKDEUtils::command( command ) ? NS_OK : NS_ERROR_FAILURE;
    }

NS_IMETHODIMP
nsKDEShellService::GetShouldCheckDefaultBrowser(PRBool* aResult)
{
  // If we've already checked, the browser has been started and this is a 
  // new window open, and we don't want to check again.
  if (mCheckedThisSession) {
    *aResult = PR_FALSE;
    return NS_OK;
  }

  nsCOMPtr<nsIPrefBranch> prefs;
  nsCOMPtr<nsIPrefService> pserve(do_GetService(NS_PREFSERVICE_CONTRACTID));
  if (pserve)
    pserve->GetBranch("", getter_AddRefs(prefs));

  if (prefs)
    prefs->GetBoolPref(PREF_CHECKDEFAULTBROWSER, aResult);

  return NS_OK;
}

NS_IMETHODIMP
nsKDEShellService::SetShouldCheckDefaultBrowser(PRBool aShouldCheck)
{
  nsCOMPtr<nsIPrefBranch> prefs;
  nsCOMPtr<nsIPrefService> pserve(do_GetService(NS_PREFSERVICE_CONTRACTID));
  if (pserve)
    pserve->GetBranch("", getter_AddRefs(prefs));

  if (prefs)
    prefs->SetBoolPref(PREF_CHECKDEFAULTBROWSER, aShouldCheck);

  return NS_OK;
}

NS_IMETHODIMP
nsKDEShellService::SetDesktopBackground(nsIDOMElement* aElement, 
                                          PRInt32 aPosition)
    {
    return NS_ERROR_NOT_IMPLEMENTED;
    }

NS_IMETHODIMP
nsKDEShellService::GetDesktopBackgroundColor(PRUint32 *aColor)
    {
    return NS_ERROR_NOT_IMPLEMENTED;
    }

NS_IMETHODIMP
nsKDEShellService::SetDesktopBackgroundColor(PRUint32 aColor)
    {
    return NS_ERROR_NOT_IMPLEMENTED;
    }

NS_IMETHODIMP
nsKDEShellService::OpenApplication(PRInt32 aApplication)
    {
    nsCStringArray command;
    if( aApplication == APPLICATION_MAIL )
        command.AppendCString( NS_LITERAL_CSTRING( "OPENMAIL" ));
    else if( aApplication == APPLICATION_NEWS )
        command.AppendCString( NS_LITERAL_CSTRING( "OPENNEWS" ));
    else
        return NS_ERROR_NOT_IMPLEMENTED;
    return nsKDEUtils::command( command ) ? NS_OK : NS_ERROR_FAILURE;
    }

NS_IMETHODIMP
nsKDEShellService::OpenApplicationWithURI(nsILocalFile* aApplication, const nsACString& aURI)
    {
    nsCStringArray command;
    command.AppendCString( NS_LITERAL_CSTRING( "RUN" ));
    nsCString app;
    nsresult rv = aApplication->GetNativePath( app );
    NS_ENSURE_SUCCESS( rv, rv );
    command.AppendCString( app );
    command.AppendCString( aURI );
    return nsKDEUtils::command( command ) ? NS_OK : NS_ERROR_FAILURE;
    }

NS_IMETHODIMP
nsKDEShellService::GetDefaultFeedReader(nsILocalFile** _retval)
    {
    *_retval = nsnull;

    nsCStringArray command;
    command.AppendCString( NS_LITERAL_CSTRING( "GETDEFAULTFEEDREADER" ));
    nsCStringArray output;
    if( !nsKDEUtils::command( command, &output ) || output.Count() != 1 )
        return NS_ERROR_FAILURE;

    nsCString path;
    path = *output[ 0 ];
    if (path.IsEmpty())
        return NS_ERROR_FAILURE;

    nsresult rv;
    nsCOMPtr<nsILocalFile> defaultReader =
        do_CreateInstance("@mozilla.org/file/local;1", &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    rv = defaultReader->InitWithNativePath(path);
    NS_ENSURE_SUCCESS(rv, rv);

    PRBool exists;
    rv = defaultReader->Exists(&exists);
    NS_ENSURE_SUCCESS(rv, rv);
    if (!exists)
        return NS_ERROR_FAILURE;

    NS_ADDREF(*_retval = defaultReader);
    return NS_OK;
    }
