/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Unix Native App Support.
 *
 * The Initial Developer of the Original Code is
 * Mozilla Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsKDEUtils.h"
#include "nsIWidget.h"

#include <gtk/gtk.h>

#include <limits.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>
#include <X11/Xlib.h>

//#define DEBUG_KDE
#ifdef DEBUG_KDE
#define KMOZILLAHELPER "kmozillahelper"
#else
// not need for lib64, it's a binary
#define KMOZILLAHELPER "/usr/lib/mozilla/kmozillahelper"
#endif

#define KMOZILLAHELPER_VERSION 6
#define MAKE_STR2( n ) #n
#define MAKE_STR( n ) MAKE_STR2( n )

static bool getKdeSession()
    {
    Display* dpy = XOpenDisplay( NULL );
    if( dpy == NULL )
        return false;
    Atom kde_full_session = XInternAtom( dpy, "KDE_FULL_SESSION", True );
    bool kde = false;
    if( kde_full_session != None )
        {
        int cnt;
        if( Atom* props = XListProperties( dpy, DefaultRootWindow( dpy ), &cnt ))
            {
            for( int i = 0;
                 i < cnt;
                 ++i )
                {
                if( props[ i ] == kde_full_session )
                    {
                    kde = true;
#ifdef DEBUG_KDE
                    fprintf( stderr, "KDE SESSION %d\n", kde );
#endif
                    break;
                    }
                }
            XFree( props );
            }
        }
    XCloseDisplay( dpy );
    return kde;
    }

static bool getKdeSupport()
    {
    nsCStringArray command;
    command.AppendCString( NS_LITERAL_CSTRING( "CHECK" ));
    command.AppendCString( NS_LITERAL_CSTRING( MAKE_STR( KMOZILLAHELPER_VERSION )));
    bool kde = nsKDEUtils::command( command );
#ifdef DEBUG_KDE
    fprintf( stderr, "KDE RUNNING %d\n", kde );
#endif
    return kde;
    }

nsKDEUtils::nsKDEUtils()
    : commandFile( NULL )
    , replyFile( NULL )
    {
    }

nsKDEUtils::~nsKDEUtils()
    {
//    closeHelper(); not actually useful, exiting will close the fd too
    }

nsKDEUtils* nsKDEUtils::self()
    {
    static nsKDEUtils s;
    return &s;
    }

static bool helperRunning = false;
static bool helperFailed = false;

bool nsKDEUtils::kdeSession()
    {
    static bool session = getKdeSession();
    return session;
    }

bool nsKDEUtils::kdeSupport()
    {
    static bool support = kdeSession() && getKdeSupport();
    return support && helperRunning;
    }

struct nsKDECommandData
    {
    FILE* file;
    nsCStringArray* output;
    GMainLoop* loop;
    bool success;
    };

static gboolean kdeReadFunc( GIOChannel*, GIOCondition, gpointer data )
    {
    nsKDECommandData* p = static_cast< nsKDECommandData* >( data );
    char buf[ 8192 ]; // TODO big enough
    bool command_done = false;
    bool command_failed = false;
    while( !command_done && !command_failed && fgets( buf, 8192, p->file ) != NULL )
        { // TODO what if the kernel splits a line into two chunks?
//#ifdef DEBUG_KDE
//        fprintf( stderr, "READ: %s %d\n", buf, feof( p->file ));
//#endif
        if( char* eol = strchr( buf, '\n' ))
            *eol = '\0';
        command_done = ( strcmp( buf, "\\1" ) == 0 );
        command_failed = ( strcmp( buf, "\\0" ) == 0 );
        nsCAutoString line( buf );
        line.ReplaceSubstring( "\\n", "\n" );
        line.ReplaceSubstring( "\\" "\\", "\\" ); //  \\ -> \ , i.e. unescape
        if( p->output && !( command_done || command_failed ))
            p->output->AppendCString( nsCString( buf )); // TODO utf8?
        }
    bool quit = false;
    if( feof( p->file ) || command_failed )
        {
        quit = true;
        p->success = false;
        }
    if( command_done )
        { // reading one reply finished
        quit = true;
        p->success = true;
        }
    if( quit )
        {
        if( p->loop )
            g_main_loop_quit( p->loop );
        return FALSE;
        }
    return TRUE;
    }

bool nsKDEUtils::command( const nsCStringArray& command, nsCStringArray* output )
    {
    return self()->internalCommand( command, NULL, false, output );
    }

bool nsKDEUtils::commandBlockUi( const nsCStringArray& command, const GtkWindow* parent, nsCStringArray* output )
    {
    return self()->internalCommand( command, parent, true, output );
    }

bool nsKDEUtils::internalCommand( const nsCStringArray& command, const GtkWindow* parent, bool blockUi,
    nsCStringArray* output )
    {
    if( !startHelper())
        return false;
    feedCommand( command );
    // do not store the data in 'this' but in extra structure, just in case there
    // is reentrancy (can there be? the event loop is re-entered)
    nsKDECommandData data;
    data.file = replyFile;
    data.output = output;
    data.success = false;
    if( blockUi )
        {
        data.loop = g_main_loop_new( NULL, FALSE );
        GtkWidget* window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
        if( parent && parent->group )
            gtk_window_group_add_window( parent->group, GTK_WINDOW( window ));
        gtk_widget_realize( window );
        gtk_widget_set_sensitive( window, TRUE );
        gtk_grab_add( window );
        GIOChannel* channel = g_io_channel_unix_new( fileno( data.file ));
        g_io_add_watch( channel, static_cast< GIOCondition >( G_IO_IN | G_IO_ERR | G_IO_HUP ), kdeReadFunc, &data );
        g_io_channel_unref( channel );
        g_main_loop_run( data.loop );
        g_main_loop_unref( data.loop );
        gtk_grab_remove( window );
        gtk_widget_destroy( window );
        }
    else
        {
        data.loop = NULL;
        while( kdeReadFunc( NULL, static_cast< GIOCondition >( 0 ), &data ))
            ;
        }
    return data.success;
    }

bool nsKDEUtils::startHelper()
    {
    if( helperRunning )
        return true;
    if( helperFailed )
        return false;
    helperFailed = true;
    int fdcommand[ 2 ];
    int fdreply[ 2 ];
    if( pipe( fdcommand ) < 0 )
        return false;
    if( pipe( fdreply ) < 0 )
        {
        close( fdcommand[ 0 ] );
        close( fdcommand[ 1 ] );
        return false;
        }
    char* args[ 2 ] = { const_cast< char* >( KMOZILLAHELPER ), NULL };
    switch( fork())
        {
        case -1:
            {
            close( fdcommand[ 0 ] );
            close( fdcommand[ 1 ] );
            close( fdreply[ 0 ] );
            close( fdreply[ 1 ] );
            return false;
            }
        case 0: // child
            {
            if( dup2( fdcommand[ 0 ], STDIN_FILENO ) < 0 )
                _exit( 1 );
            if( dup2( fdreply[ 1 ], STDOUT_FILENO ) < 0 )
                _exit( 1 );
            int maxfd = 1024; // close all other fds
            struct rlimit rl;
            if( getrlimit( RLIMIT_NOFILE, &rl ) == 0 )
                maxfd = rl.rlim_max;
            for( int i = 3;
                 i < maxfd;
                 ++i )
                close( i );
#ifdef DEBUG_KDE
            execvp( KMOZILLAHELPER, args );
#else
            execv( KMOZILLAHELPER, args );
#endif
            _exit( 1 ); // failed
            }
        default: // parent
            {
            commandFile = fdopen( fdcommand[ 1 ], "w" );
            replyFile = fdopen( fdreply[ 0 ], "r" );
            close( fdcommand[ 0 ] );
            close( fdreply[ 1 ] );
            if( commandFile == NULL || replyFile == NULL )
                {
                closeHelper();
                return false;
                }
            // ok, helper ready, getKdeRunning() will check if it works
            }
        }
    helperFailed = false;
    helperRunning = true;
    return true;
    }

void nsKDEUtils::closeHelper()
    {
    if( commandFile != NULL )
        fclose( commandFile ); // this will also make the helper quit
    if( replyFile != NULL )
        fclose( replyFile );
    helperRunning = false;
    }

void nsKDEUtils::feedCommand( const nsCStringArray& command )
    {
    for( int i = 0;
         i < command.Count();
         ++i )
        {
        nsCString line = *command[ i ];
        line.ReplaceSubstring( "\\", "\\" "\\" ); // \ -> \\ , i.e. escape
        line.ReplaceSubstring( "\n", "\\n" );
#ifdef DEBUG_KDE
        fprintf( stderr, "COMM: %s\n", line.get());
#endif
        fputs( line.get(), commandFile );
        fputs( "\n", commandFile );
        }
    fputs( "\\E\n", commandFile ); // done as \E, so it cannot happen in normal data
    fflush( commandFile );
    }
